/*
 * USB SQcam WebCam driver based on the ViCam driver written by Joe Burks, 
 * Christopher L Cheney, Pavel Machek, John Tyner, Monroe Williams.
 *
 * Thanks to Theodore Kilgore for his help.
 * Ported to kernel 2.6 by Wilfred Nelson
 * 
 * Modified by Philippe Corbes to:
 * - Improve plug(sqcam_probe)/unplug(sqcam_disconnect)
 * - change deprecated code on 2.6, and suppress usbvideo.h include
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * This source code is based heavily on the CPiA webcam driver which was
 * written by Peter Pregler, Scott J. Bertin and Johannes Erdfelt
 *
 * Portions of this code were also copied from usbvideo.c
 *
 * The bayer decoding functions are copied from libgphoto2
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/videodev.h>
#include <linux/usb.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>

#include "image_filter.c"

// #define SQCAM_DEBUG

#ifndef MODULE_LICENSE
#define MODULE_LICENSE(a)
#endif

#ifndef bool
#define bool int
#endif

#ifdef SQCAM_DEBUG
#define ADBG(lineno,fmt,args...) printk(fmt, jiffies, __FUNCTION__, lineno, ##args)
#define DBG(fmt,args...) ADBG((__LINE__),KERN_DEBUG __FILE__"(%ld):%s (%d):"fmt,##args)
#else
#define DBG(fmn,args...) do {} while(0)
#endif

/* Version Information */
#define DRIVER_VERSION "v0.2"
#define DRIVER_AUTHOR "Marcell Lengyel, miketkf@mailbox.hu"
#define DRIVER_DESC "SQcam WebCam Driver"

/* Define these values to match your device */
#define USB_SQCAM_VENDOR_ID		0x2770
#define USB_SQCAM_PRODUCT_ID	0x9120

#define SQCAM_BYTES_PER_PIXEL	3
#define SQCAM_MAX_READ_SIZE		0x8000
#define SQCAM_MAX_RAW_SIZE		320*240+64
#define SQCAM_MAX_FRAME_SIZE	(SQCAM_BYTES_PER_PIXEL*320*240)
#define SQCAM_FRAMES	2
#define SHARPNESS_MIN	0
#define SHARPNESS_MAX	6
#define FRAMERATE_MIN	0
#define FRAMERATE_MAX	6

#define MAX_SQCAM 4


/************************ Camera Descriptor  **************************/
struct sqcam_camera {

	struct video_device *vdev;	// v4l video device
	struct usb_device   *udev;	// usb device

	u8 *raw_image;				// raw data captured from the camera
	u8 *framebuf;				// processed data in RGB24 format
/*	unsigned char gamma_table[256]; */

	struct image_filter *filter;	// gamma, red, green, blue
	
	struct semaphore busy_lock;	// guard against SMP multithreading

	int channel;
	bool is_initialized;
	bool is_removed ;
	bool is_opened ;
	u8 bulkEndpoint;
	u32 framebuf_size;			// # of valid bytes in framebuf
	u32 framebuf_read_start;	// position in frame buf that a read is happening at.

	__u8 iface;

};


/* Video device number (-1 is first available) */
static int video_nr = -1;


#define	SQCAM_T(uvd)	((struct sqcam_camera *)((uvd)->user_data))

static int sqcam_probe( struct usb_interface *intf, const struct usb_device_id *id);
static void sqcam_disconnect(struct usb_interface *intf);
static void read_frame(struct sqcam_camera *cam, int framenum);

static int gp_bayer_decode (unsigned char *input, int w, int h,
			     unsigned char *output, int tile);

static void usbvideo_rvfree(void *mem, unsigned long size);
static void *usbvideo_rvmalloc(unsigned long size);
unsigned long usbvideo_kvirt_to_pa(unsigned long adr);

			  
/*static int flags  = FLAGS_DISPLAY_HINTS | FLAGS_OVERLAY_STATS; */
static const int min_canvasWidth  = 8;
static const int min_canvasHeight = 4;


static int
send_control_msg(int set, struct sqcam_camera *cam, u8 request, u16 value,
		 u16 index, unsigned char *cp, u16 size)
{
	int status;

	// for reasons not yet known to me, you can't send USB control messages
	// with data in the module (if you are compiled as a module).  Whatever
	// the reason, copying it to memory allocated as kernel memory then
	// doing the usb control message fixes the problem.

	struct usb_device *udev = cam->udev;
	unsigned char *transfer_buffer = kmalloc(size, GFP_KERNEL);

	memcpy(transfer_buffer, cp, size);
	/*down(&cam->busy_lock);*/

	status = usb_control_msg(udev,
				 set ? usb_sndctrlpipe(udev, 0) : usb_rcvctrlpipe(udev, 0),
				 request,
				 (set ? USB_DIR_OUT : USB_DIR_IN) | USB_TYPE_VENDOR |
				 USB_RECIP_DEVICE, value, index,
				 transfer_buffer, size, HZ);

	kfree(transfer_buffer);

	DBG("ctrl msg:: value: %0x, index: %0x, size: %i\n", value, index, size);
	status = min(status, 0);
	if (status < 0) 
	{
		printk(KERN_ERR "send_msg: Failed sending control message, error %d.\n", status);
		printk(KERN_ERR "ctrl msg:: dir: %i, value: %0x, index: %0x, size: %i\n", set, value, index,  size);
	}
	/*up(&cam->busy_lock);*/

	return status;
}

static int
initialize_camera(struct sqcam_camera *cam)
{
	int status;
	unsigned char ch[5], zh;
	
	/*
000001	CS        80 06 00 01 00 00 12 00	

	C<  0000:  1201 1001 ffff ff08 7027 2091 0001 0002  ........p' .....
	    0010:  0001                                     ..

000002	CS        80 06 00 02 00 00 40 00	

	C<  0000:  0902 2700 0101 0080 fa09 0400 0003 ffff  ..'.............
	    0010:  ff00 0705 8102 4000 0007 0502 0240 0000  ......@......@..
	    0020:  0705 8303 0100 03                        .......	
	*/
	
	
	zh = 0x00;
	if ((status = send_control_msg(1, cam, 0x0c, 0x06, 0xf0, &zh, 1)) < 0)
	{
		printk(KERN_ERR "initalize: message 1, error %d.\n",status);
		return status;
	}
	zh = 0x00;
	if ((status = send_control_msg(0, cam, 0x0c, 0x07, 0, &zh, 1)) < 0)
	{
		printk(KERN_ERR "initalize: message 2, error %d.\n",status);
		return status;
	}
	
/*

000005	CS        c0 0c 07 00 00 00 01 00	

	C<  0000:  09                                       .

000006	CS        c0 0c 07 00 00 00 01 00	

	C<  0000:  05                                       .

000007	CS        c0 0c 07 00 00 00 01 00	

	C<  0000:  00                                       .

000008	CS        c0 0c 07 00 00 00 01 00	

	C<  0000:  01                                       .

000009	CS        40 0c 06 00 a0 00 01 00	

	C>  0000:  50                                       P

000010	CS        c0 0c 07 00 00 00 01 00	

	C<  0000:  50                                       P


*/	
	
	ch[0] = 0x00;
	if ((status = send_control_msg(0,cam, 0x0c, 0x07, 0, &ch[0], 4)) < 0)
	{
		printk(KERN_ERR "initalize: message 3, error %d.\n",status);
		return status;
	}
	zh = 0x00;
	if ((status = send_control_msg(1, cam, 0x0c, 0x06, 0xa0, &zh, 1)) < 0)
	{
		printk(KERN_ERR "initalize: message 4, error %d.\n",status);
		return status;
	}
	zh = 0x00;
	if ((status = send_control_msg(0, cam, 0x0c, 0x07, 0, &zh, 1)) < 0)
	{
		printk(KERN_ERR "initalize: message, error %d.\n",status);
		return status;
	}
    cam->channel = 1;
	return 0;
}

static int
reset_camera(struct sqcam_camera *cam)
{
	int status;
	unsigned char ch;

	ch = 0xff;
	if ((status = send_control_msg(1, cam, 0x0c, 0xc0, 0x00, &ch, 1)) < 0)
		return status;
	if ((status = send_control_msg(1, cam, 0x0c, 0x06, 0xa0, NULL, 0)) < 0)
		return status;
	if ((status = send_control_msg(0, cam, 0x0c, 0x07, 0, &ch, 1)) < 0)
		return status;
	return 0;
}

static int
sqcam_ioctl(struct inode *inode, struct file *file, unsigned int ioctlnr, unsigned long ul_arg)
{
	void *arg = (void *)ul_arg;
	struct video_device *dev = file->private_data;
	struct sqcam_camera *cam = video_get_drvdata(dev);
	int retval = 0;

	DBG("sqcam_ioctl\n");

	if (!cam)
		return -ENODEV;
	if (!cam->udev)
		return -EIO;

	/* make this _really_ smp-safe */
	/*if (down_interruptible(&cam->busy_lock))
		return -EINTR;*/

	switch (ioctlnr) {

	/* query capabilites */
	case VIDIOCGCAP:
		{
			struct video_capability b;

			DBG("VIDIOCGCAP\n");
			strcpy(b.name, "SQcam-based Camera");
			b.type = VID_TYPE_CAPTURE;
			b.channels = 1;
			b.audios = 0;
			b.maxwidth = 320;	/* VIDEOSIZE_CIF */
			b.maxheight = 240;
			b.minwidth = 320;	/* VIDEOSIZE_160_120 */
			b.minheight = 240;

			if (copy_to_user(arg, &b, sizeof (b)))
				retval = -EFAULT;

			break;
		}
	/* get/set video source - we are a camera and nothing else */
	case VIDIOCGCHAN:
		{
			struct video_channel v;

			DBG("VIDIOCGCHAN\n");
			if (copy_from_user(&v, arg, sizeof (v))) {
				retval = -EFAULT;
				break;
			}
			if (v.channel != 0 ) {
				retval = -EINVAL;
				break;
			}

			/* v.channel = 0;*/
            cam->channel = v.channel;
			strcpy(v.name, "Camera");
			v.tuners = 0;
			v.flags = 0;
			v.type = VIDEO_TYPE_CAMERA;
			v.norm = 0;

			if (copy_to_user(arg, &v, sizeof (v)))
				retval = -EFAULT;
			break;
		}

	case VIDIOCSCHAN:
		{
			int v;

			if (copy_from_user(&v, arg, sizeof (v)))
				retval = -EFAULT;
			DBG("VIDIOCSCHAN %d\n", v);

			if (retval == 0 && v != 0)
				retval = -EINVAL;
            cam->channel = v;
			break;
		}

	/* image properties */
	case VIDIOCGPICT:
		{
			struct video_picture vp;
			DBG("VIDIOCGPICT\n");

			memset(&vp, 0, sizeof (struct video_picture));

			vp.depth = 24;
			vp.palette = VIDEO_PALETTE_RGB24;
			vp.colour = cam->filter->Gamma << 13;
			if (copy_to_user (arg, &vp, sizeof (struct video_picture)))
				retval = -EFAULT;
			break;
		}

	case VIDIOCSPICT:
		{
			struct video_picture vp;
			int gamma;
			
			if(copy_from_user(&vp, (struct video_picture *) arg, sizeof(struct video_picture)))
			{
				printk(KERN_ERR "VIDIOCSPICT error!\n");
				retval = -EFAULT;
			}
			else
			{
				DBG("VIDIOCSPICT depth:%d, palette:%d, hue:%d, colour:%d, whiteness:%d, brightness:%d, contrast:%d\n", 
								vp.depth, vp.palette, vp.hue, vp.colour, vp.whiteness, vp.brightness, vp.contrast);

				if (vp.depth != 24 || vp.palette != VIDEO_PALETTE_RGB24)
					retval = -EINVAL;

				/* Convert colour parameter to gamma parameter */
				gamma = (int)(vp.colour >> 13);
				if (gamma > 6) gamma = 6;
				if (cam->filter->Gamma != gamma) {
					cam->filter->Gamma = gamma;
					cam->filter->change = FILTER_UPDATE_GAMMA;
					image_filter_update(cam->filter);
				}
			}

			break;
		}

	/* get/set capture window */
	case VIDIOCGWIN:
		{
			struct video_window vw;
			vw.x = 0;
			vw.y = 0;
			vw.width = 320;
			vw.height = 240;
			vw.chromakey = 0;
			vw.flags = 0;
			vw.clips = NULL;
			vw.clipcount = 0;

			DBG("VIDIOCGWIN\n");

			if (copy_to_user
			    ((void *) arg, (void *) &vw, sizeof (vw)))
				retval = -EFAULT;

			// I'm not sure what the deal with a capture window is, it is very poorly described
			// in the doc.  So I won't support it now.
			break;
		}

	case VIDIOCSWIN:
		{

			struct video_window *vw = (struct video_window *) arg;
			DBG("VIDIOCSWIN %d x %d\n", vw->width, vw->height);

			if ( vw->width != 320 || vw->height != 240 )
				retval = -EFAULT;
			
			break;
		}

	/* mmap interface */
	case VIDIOCGMBUF:
		{
			struct video_mbuf vm;
			int i;

			DBG("VIDIOCGMBUF\n");
			memset(&vm, 0, sizeof (vm));
			vm.size = SQCAM_MAX_FRAME_SIZE * SQCAM_FRAMES;
			vm.frames = SQCAM_FRAMES;
			for (i = 0; i < SQCAM_FRAMES; i++)
				vm.offsets[i] = SQCAM_MAX_FRAME_SIZE * i;

			if (copy_to_user ((void *) arg, (void *) &vm, sizeof (vm)))
				retval = -EFAULT;

			break;
		}

	case VIDIOCMCAPTURE:
		{
			struct video_mmap vm;
			// int video_size;

			if (copy_from_user
			    ((void *) &vm, (void *) arg, sizeof (vm))) {
				retval = -EFAULT;
				break;
			}

			DBG("VIDIOCMCAPTURE frame=%d, height=%d, width=%d, format=%d.\n",vm.frame,vm.width,vm.height,vm.format);

			if ( vm.frame >= SQCAM_FRAMES || vm.format != VIDEO_PALETTE_RGB24 )
				retval = -EINVAL;

			// in theory right here we'd start the image capturing
			// (fill in a bulk urb and submit it asynchronously)
			//
			// Instead we're going to do a total hack job for now and
			// retrieve the frame in VIDIOCSYNC

			break;
		}

	case VIDIOCSYNC:
		{
			int frame;

			if (copy_from_user((void *) &frame, arg, sizeof (int))) {
				retval = -EFAULT;
				break;
			}
			DBG("VIDIOCSYNC: %d\n", frame);
			read_frame(cam, frame);

			break;
		}

	/* pointless to implement overlay with this camera */
	case VIDIOCCAPTURE:
	case VIDIOCGFBUF:
	case VIDIOCSFBUF:
	case VIDIOCKEY:
		retval = -EINVAL;
		break;

	/* tuner interface - we have none */
	case VIDIOCGTUNER:
	case VIDIOCSTUNER:
	case VIDIOCGFREQ:
	case VIDIOCSFREQ:
		retval = -EINVAL;
		break;

	/* audio interface - we have none */
	case VIDIOCGAUDIO:
	case VIDIOCSAUDIO:
		retval = -EINVAL;
		break;

	default:
		retval = -ENOIOCTLCMD;
		break;
	}

	/*up(&cam->busy_lock);*/
	return retval;
}

static int
sqcam_open(struct inode *inode, struct file *file)
{
    struct video_device *dev = video_devdata(file);
	struct sqcam_camera *cam = video_get_drvdata(dev);
	unsigned char siz;
	int n;
	int intr;

	DBG("sqcam_open\n");

	if (!cam) 
	{
		printk(KERN_ERR "sqcam video_device improperly initialized");
    	return -EIO;
    }

	intr = down_interruptible(&cam->busy_lock);
	if (intr)
	{
		printk(KERN_ERR "sqcam_open error on lock access");
		return -EINTR;
	}

	if (cam->is_opened) 
	{
		printk(KERN_ERR "sqcam_open called on already opened camera");
		up(&cam->busy_lock);
		return -EBUSY;
	}

	cam->raw_image = kmalloc(SQCAM_MAX_RAW_SIZE, GFP_KERNEL);
	if (!cam->raw_image) 
	{
		if (!cam->raw_image) 
		{
			printk(KERN_ERR "sqcam_open error on raw_image allocation");
			up(&cam->busy_lock);
			return -ENOMEM;
		}
	}

	cam->framebuf = usbvideo_rvmalloc(SQCAM_MAX_FRAME_SIZE * SQCAM_FRAMES);
	if (!cam->framebuf) 
	{
		if (!cam->framebuf) 
		{
			printk(KERN_ERR "sqcam_open error on framebuf allocation");
			kfree(cam->raw_image);
			up(&cam->busy_lock);
			return -ENOMEM;
		}
	}

	n = 0;
	if (!cam->is_initialized) 
	{
		n = initialize_camera(cam);
	}

	if (n >= 0)
	{
		siz = 0x00;
		n = send_control_msg(1, cam, 0x0c, 0xc0, 0, &siz, 0);
		if (n < 0)
			printk(KERN_ERR "sqcam_open: message 1, error %d.\n",n);
	}

	if (n >= 0) 
	{
		siz = 0x61; // 0x60 -> 160x120, 0x61 -> 320x240
		n = send_control_msg(1, cam, 0x0c, 0x06, siz, NULL, 0);
		if (n < 0)
			printk(KERN_ERR "sqcam_open: mode 320x240, error %d.\n",n);
	}

	if (n >= 0) 
	{
		siz = 0x00;
		n = send_control_msg(0, cam, 0x0c, 0x07, 0x00, &siz, 1);
		if (n < 0)
			printk(KERN_ERR "sqcam_open: message 3, error %d.\n",n);
	}
 
 	image_filter_set(cam->filter);

	file->private_data = dev;
 	
 	up(&cam->busy_lock);

    if (n >= 0) 
    {
		DBG("sqcam_open end.\n");
		cam->is_initialized = 1;
		cam->is_opened = 1;
		return 0;
    } else {
		printk(KERN_ERR "sqcam_open: -EIO error\n");
	  	cam->is_initialized = 0;
    	return -EIO;
    }

}

static int 
sqcam_close(struct inode *inode, struct file *file)
{
	struct video_device *dev = file->private_data;
	struct sqcam_camera *cam = video_get_drvdata(dev);
	struct usb_device *udev;

	DBG("sqcam_close\n");

	/* it's not the end of the world if
	 * we fail to turn the camera off.
	 */

	kfree(cam->raw_image);
	usbvideo_rvfree(cam->framebuf, SQCAM_MAX_FRAME_SIZE * SQCAM_FRAMES);

	down(&cam->busy_lock);
	cam->is_opened = 0;
	reset_camera(cam);
	udev = cam->udev;
	up(&cam->busy_lock);

	file->private_data = NULL;

	if (!cam->is_opened && !udev)
		kfree(cam);

	return 0;
}


#define DATA_HEADER_SIZE 64

void sqcam_decode_bayer(const u8 *data, u8 *rgb, struct sqcam_camera *cam)
{
	int j;
	unsigned char b;
	unsigned char *src;
	unsigned char *dst;

	src = (u8 *) data + DATA_HEADER_SIZE;
	dst = rgb;

    /* Reverse the data, else the pic is upside down. */
    for (j = 0; j < 320*240 / 2; j++) 
    {
	    b = src[j];
	    src[j] = src[320*240 - 1 - j];
	    src[320*240 - 1 - j] = b;
    } 
	gp_bayer_decode(src, 320, 240, dst, cam->channel);

    /* Apply filter if necessary */
    if (cam->filter->enabled)
    	image_filter_apply(cam->filter, dst  , 320*240);
}

static void
read_frame(struct sqcam_camera *cam, int framenum)
{
	unsigned char siz, msg;
	int n, read;
	int actual_length;

	if (!cam->udev)
		return;

	msg = 0x03; 
	read = 320 * 240 + 64;
	// read 'read' bytes in SQCAM_MAX_READ_SIZE chunks
		
	do {
		siz = 0x50;
		n = send_control_msg(1, cam, 0x0c,0x03, 
				(read > SQCAM_MAX_READ_SIZE) ? SQCAM_MAX_READ_SIZE : read, 
				&siz, 1);
		if (n < 0)
		{
			printk(KERN_ERR " sqcam_read_frame: Problem sending frame capture control message");
			goto done;
		}
		n = usb_bulk_msg(cam->udev,
			 usb_rcvbulkpipe(cam->udev, cam->bulkEndpoint),
			 cam->raw_image + (320*240+64 - read),
			 (read > SQCAM_MAX_READ_SIZE) ? SQCAM_MAX_READ_SIZE : read,
			 &actual_length, 500);
		read -= actual_length;
		DBG("blk msg:: size: %d\n", actual_length);
		if (n < 0)
		{
			printk(KERN_ERR "Problem during bulk read of frame data: %d\n", n);
			read = 0;
		}
	} while (read);	
	
	sqcam_decode_bayer(cam->raw_image,
			 cam->framebuf + framenum * SQCAM_MAX_FRAME_SIZE,
             cam );
	cam->framebuf_size = 320 * 240 * SQCAM_BYTES_PER_PIXEL;
	cam->framebuf_read_start = 0;
  done: 
    DBG("sqcam_read_frame: exiting \n");
}

static ssize_t
sqcam_read( struct file *file, char *buf, size_t count, loff_t *ppos )
{
	struct video_device *dev = file->private_data;
	struct sqcam_camera *cam = video_get_drvdata(dev);
	int intr;
	
	DBG("sqcam_read %d bytes.\n", (int) count);
	
	intr = down_interruptible(&cam->busy_lock);
	if (intr)
		return -EINTR;

	if (*ppos >= SQCAM_MAX_FRAME_SIZE) 
	{
		*ppos = 0;
		return 0;
	}

	if (*ppos == 0)
		read_frame(cam, 0);

	count = min_t(size_t, count, SQCAM_MAX_FRAME_SIZE - *ppos);

	if (copy_to_user(buf, &cam->framebuf[*ppos], count)) 
	{
		count = -EFAULT;
	} else {
		*ppos += count;
	}

	if (count == SQCAM_MAX_FRAME_SIZE)
		*ppos = 0;

	DBG("read %d bytes. finished\n", (int) count);

	up(&cam->busy_lock);
	return count;
}


static int
sqcam_mmap(struct file *file, struct vm_area_struct *vma)
{
	// TODO: allocate the raw frame buffer if necessary
	unsigned long start;
	unsigned long size;
	unsigned long page, pos;
	struct video_device *dev;
	struct sqcam_camera *cam;

	start = vma->vm_start;
	size  = vma->vm_end-vma->vm_start;
	dev = file->private_data;

	DBG("sqcam_mmap: %ld\n", size);

	/* We let mmap allocate as much as it wants because Linux was adding 2048 bytes
	 * to the size the application requested for mmap and it was screwing apps up.
	 if (size > SQCAM_FRAMES*SQCAM_MAX_FRAME_SIZE)
	 return -EINVAL;
	 */
	
	cam = video_get_drvdata(dev);
	if (!cam)
		return -ENODEV;

        
	if (down_interruptible(&cam->busy_lock))
		return -EINTR;


	if (!cam->framebuf)
	{
		cam->framebuf =
		    usbvideo_rvmalloc(SQCAM_MAX_FRAME_SIZE * SQCAM_FRAMES);
		if (!cam->framebuf) 
		{
			up(&cam->busy_lock);
			return -ENOMEM;
		}
	}

	pos = (unsigned long) (cam->framebuf);
	while (size > 0) {
		page = usbvideo_kvirt_to_pa(pos);
// #ifdef HAS_REMAP_PAGE_RANGE
// 		if (remap_page_range(vma, start, page, PAGE_SIZE,
//                      PAGE_SHARED)) {
// #else
		if (remap_pfn_range(vma, start, page >> PAGE_SHIFT, PAGE_SIZE,
                    PAGE_SHARED)) 
        {
// #endif
			up(&cam->busy_lock);
			return -EAGAIN;
		}
		start += PAGE_SIZE;
		pos += PAGE_SIZE;
		if (size > PAGE_SIZE)
			size -= PAGE_SIZE;
		else
			size = 0;
	}

	up(&cam->busy_lock);

	return 0;
}



/****************************************************************************
*  sysfs
***************************************************************************/
static ssize_t
show_pictsetting(struct class_device *cd, char *buf)
{
	struct video_device *vdev   = to_video_device(cd);
	struct sqcam_camera *cam    = video_get_drvdata(vdev);
	struct image_filter *filter = cam->filter;
	return snprintf(buf, 128,
			"Gamma=%d, OffsetRed=%d, OffsetBlue=%d, OffsetGreen=%d, GainRed=%d, GainBlue=%d, GainGreen= %d \n",
			filter->Gamma, 
			filter->OffsetRed, filter->OffsetBlue, filter->OffsetGreen, 
			filter->GainRed, filter->GainBlue, filter->GainGreen);
}

static CLASS_DEVICE_ATTR(pictsetting, S_IRUGO, show_pictsetting, NULL);

static void
sqcam_create_sysfs(struct video_device *vdev)
{
	video_device_create_file(vdev, &class_device_attr_pictsetting);
}



/* table of devices that work with this driver */
static struct usb_device_id sqcam_table[] = {
	{USB_DEVICE(USB_SQCAM_VENDOR_ID, USB_SQCAM_PRODUCT_ID)},
	{}			/* Terminating entry */
};

MODULE_DEVICE_TABLE(usb, sqcam_table);

static struct file_operations sqcam_fops = {
	.owner =  THIS_MODULE,
	.open =   sqcam_open,
	.release = sqcam_close,
	.read =   sqcam_read,
	.mmap =  sqcam_mmap,
	.ioctl = sqcam_ioctl,
	.llseek = no_llseek,
};

static struct usb_driver sqcam_driver = {
//        .owner          = THIS_MODULE,
        .name           = "sqcam",
        .probe          = sqcam_probe,
        .disconnect     = sqcam_disconnect,
        .id_table       = sqcam_table,
};



static struct video_device sqcam_template = {
	.owner = THIS_MODULE,
	.name  = "SQcam-based USB Camera",
	.type   = VID_TYPE_CAPTURE,
	.hardware = VID_HARDWARE_SE401, /* need an own value */
	.release = &video_device_release,
    .fops  =  &sqcam_fops,
	.minor = -1,
};

/**
 *	sqcam_probe
 *
 *	Called by the usb core when a new device is connected that it thinks
 *	this driver might be interested in.
 */
static int
sqcam_probe(struct usb_interface *intf, const struct usb_device_id *id)
{
	const struct usb_host_interface *interface;
	struct sqcam_camera *cam;
	int nas, bulkEndpoint = 0;
	const struct usb_endpoint_descriptor *endpoint;
	struct usb_device *dev = interface_to_usbdev(intf);


	/* We don't handle multi-config cameras */
	if (dev->descriptor.bNumConfigurations != 1)
		goto nodevice;

	if ((cam = kmalloc(sizeof (struct sqcam_camera), GFP_KERNEL)) == NULL) 
	{
		err("couldn't kmalloc spcam struct");
		goto error;
	}
	memset(cam, 0, sizeof (struct sqcam_camera));

	if ((cam->filter = kmalloc(sizeof (struct image_filter), GFP_KERNEL)) == NULL) 
	{
		err("couldn't kmalloc image_filter struct");
		goto error;
	}
	memset(cam->filter, 0, sizeof (struct image_filter));

	cam->iface = intf->altsetting->desc.bInterfaceNumber;
	if (cam->iface > 0)
		goto nodevice;

	cam->udev = dev;

	/* See if the device offered us matches what we can accept */
	if ((dev->descriptor.idVendor != USB_SQCAM_VENDOR_ID) ||
	    (dev->descriptor.idProduct != USB_SQCAM_PRODUCT_ID)) 
	{
		printk(KERN_WARNING " Devices not found !! ");
		goto error;
	}

 	/* Init image filters */
 	image_filter_set(cam->filter);

	printk(KERN_INFO "SQcam based webcam connected\n");

	/* Initialise the camera bridge */
	nas = dev->actconfig->interface[cam->iface]->num_altsetting;
	if (nas != 1) 
	{
		printk(KERN_WARNING "Expected only one alternate setting for this camera!\n");
		return -EINVAL;
	}

	interface = &dev->actconfig->interface[cam->iface]->altsetting[0];
	DBG(KERN_DEBUG "Interface %d. has %u. endpoints!\n",
	       interface->desc.bInterfaceNumber, 
	       (unsigned) (interface->desc.bNumEndpoints));
	endpoint = &interface->endpoint[0].desc;

	if ((endpoint->bEndpointAddress & 0x80) &&
	    ((endpoint->bmAttributes & 3) == 0x02)) 
	{
		/* we found a bulk in endpoint */
		bulkEndpoint = endpoint->bEndpointAddress;
	} else {
		printk(KERN_ERR "No bulk in endpoint was found ?! (this is bad)\n");
	}

	init_MUTEX(&cam->busy_lock);
	cam->bulkEndpoint = bulkEndpoint;
	cam->is_initialized = 0;
	cam->is_opened = 0;

	/* Init video stuff */
	cam->vdev = video_device_alloc();
	if (!cam->vdev)
		goto error;

	memcpy(cam->vdev, &sqcam_template, sizeof (sqcam_template));
	cam->vdev->dev = &dev->dev;
	video_set_drvdata(cam->vdev, cam);
	printk(KERN_DEBUG "Setting video device = %p, cam = %p\n", cam->vdev, cam);
	if (video_register_device(cam->vdev, VFL_TYPE_GRABBER, video_nr) < 0) 
	{
		printk(KERN_WARNING "video_register_device failed");
		goto error;
	}

	printk(KERN_INFO "%s %s now controlling video device %d.\n",
			DRIVER_DESC, DRIVER_VERSION, cam->vdev->minor);

	usb_set_intfdata (intf, cam);
	sqcam_create_sysfs(cam->vdev);
	return 0;

  error:
	if (cam->vdev) {
		if (cam->vdev->minor == -1)
			video_device_release(cam->vdev);
		else
			video_unregister_device(cam->vdev);
		cam->vdev = NULL;
	}
	if (cam) {
		kfree(cam);
		cam = NULL;
	}
	return -EIO;
  nodevice:
	return -ENODEV;
}

static void
sqcam_disconnect(struct usb_interface *intf)
{
    struct sqcam_camera *cam = usb_get_intfdata(intf);

    if (!cam)
    {
		printk(KERN_DEBUG "SQCam-based WebCam is already disconnected!!\n");
    	return;
    }

    down(&cam->busy_lock);
	dev_set_drvdata(&intf->dev, NULL);
	if (cam->vdev) 
	{
		DBG(KERN_DEBUG "SQCam-based WebCam unregister device.\n");
		video_unregister_device(cam->vdev);
	}
	cam->udev = NULL;
	up(&cam->busy_lock);

	/* the only thing left to do is synchronize with
	 * our close/release function on who should release
	 * the camera memory. if there are any users using the
	 * camera, it's their job. if there are no users,
	 * it's ours.
	 */

	if (cam && !cam->is_opened) 
	{
		DBG(KERN_DEBUG "SQCam-based WebCam freeing cam.\n");
		if (cam->filter) 
			kfree(cam->filter);
		kfree(cam);
		cam = NULL;
	}
	printk(KERN_DEBUG "SQCam-based WebCam disconnected.\n");
	
}

static int __init
usb_sqcam_init(void)
{
	int retval;
	DBG(KERN_INFO "SQcam-based WebCam driver startup\n");
	retval = usb_register(&sqcam_driver);
	if (retval)
		printk(KERN_WARNING "usb_register failed!\n");
	return retval;
}

static void __exit
usb_sqcam_exit(void)
{
	DBG(KERN_INFO "SQcam-based WebCam driver shutdown\n");
	usb_deregister(&sqcam_driver);
}

module_init(usb_sqcam_init);
module_exit(usb_sqcam_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
/* bayer.c
 *
 * Copyright  2001 Lutz Mller <urc8@rz.uni-karlsruhe.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details. 
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/* This is from libgphoto2 */


static int tile_colors[8][4] = {
	{0, 1, 1, 2},
	{1, 0, 2, 1},
	{2, 1, 1, 0},
	{1, 2, 0, 1},
	{0, 1, 1, 2},
	{1, 0, 2, 1},
	{2, 1, 1, 0},
	{1, 2, 0, 1}};

#define RED 0
#define GREEN 1
#define BLUE 2

static int
gp_bayer_expand (unsigned char *input, int w, int h, unsigned char *output,
		 int tile)
{
	int x, y, i;
	int colour, bayer;
	char *ptr = input;

	switch (tile) {

		case 0:
		case 1:
		case 2:
		case 3: 

			for (y = 0; y < h; ++y)
				for (x = 0; x < w; ++x, ++ptr) {
					bayer = (x&1?0:1) + (y&1?0:2);

					colour = tile_colors[tile][bayer];

					i = (y * w + x) * 3;

					output[i+RED]    = 0;
					output[i+GREEN]  = 0;
					output[i+BLUE]   = 0;
					output[i+colour] = *ptr;
				}
			break;

		case 4: //BAYER_TILE_RGGB_INTERLACED:
		case 5: //BAYER_TILE_GRBG_INTERLACED:
		case 6: //BAYER_TILE_BGGR_INTERLACED:
		case 7: //BAYER_TILE_GBRG_INTERLACED:
 

			for (y = 0; y < h; ++y, ptr+=w)
				for (x = 0; x < w; ++x) {
					bayer = (x&1?0:1) + (y&1?0:2);
	
					colour = tile_colors[tile][bayer];
	
					i = (y * w + x) * 3;

					output[i+RED]    = 0;
					output[i+GREEN]  = 0;
					output[i+BLUE]   = 0;
					output[i+colour] = (x&1)? ptr[x>>1]:ptr[(w>>1)+(x>>1)];
				}
			break;
	}

	return (0);
}

#define AD(x, y, w) ((y)*(w)*3+3*(x))

static int
gp_bayer_interpolate (unsigned char *image, int w, int h, int tile)
{
	int x, y, bayer;
	int p0, p1, p2, p3;
	int div, value;


	switch (tile) {
	default:
	case 0: //BAYER_TILE_RGGB:
	case 4: //BAYER_TILE_RGGB_INTERLACED:
		p0 = 0; p1 = 1; p2 = 2; p3 = 3;
		break;
	case 1: //BAYER_TILE_GRBG:
	case 5: //BAYER_TILE_GRBG_INTERLACED:
		p0 = 1; p1 = 0; p2 = 3; p3 = 2;
		break;
	case 2: //BAYER_TILE_BGGR:
	case 6: //BAYER_TILE_BGGR_INTERLACED:
		p0 = 3; p1 = 2; p2 = 1; p3 = 0;
		break;
	case 3: //BAYER_TILE_GBRG:
	case 7: //BAYER_TILE_GBRG_INTERLACED:	
		p0 = 2; p1 = 3; p2 = 0; p3 = 1;
		break;
	}


	//p0 = 3; p1 = 2; p2 = 1; p3 = 0; // 2
	//p0 = 0; p1 = 1; p2 = 2; p3 = 3;


	for (y = 0; y < h; y++)
		for (x = 0; x < w; x++) {
			bayer = (x&1?0:1) + (y&1?0:2);
			if ( bayer == p0 ) {
				
				/* red. green lrtb, blue diagonals */
				div = value = 0;
				if (y) {
					value += image[AD(x,y-1,w)+GREEN];
					div++;
				}
				if (y < (h - 1)) {
					value += image[AD(x,y+1,w)+GREEN];
					div++;
				}
				if (x) {
					value += image[AD(x-1,y,w)+GREEN];
					div++;
				}
				if (x < (w - 1)) {
					value += image[AD(x+1,y,w)+GREEN];
					div++;
				}
				image[AD(x,y,w)+GREEN] = value / div;

				div = value = 0;
				if ((y < (h - 1)) && (x < (w - 1))) {
					value += image[AD(x+1,y+1,w)+BLUE];
					div++;
				}
				if ((y) && (x)) {
					value += image[AD(x-1,y-1,w)+BLUE];
					div++;
				}
				if ((y < (h - 1)) && (x)) {
					value += image[AD(x-1,y+1,w)+BLUE];
					div++;
				}
				if ((y) && (x < (w - 1))) {
					value += image[AD(x+1,y-1,w)+BLUE];
					div++;
				}
				image[AD(x,y,w)+BLUE] = value / div;

			} else if (bayer == p1) {
				
				/* green. red lr, blue tb */
				div = value = 0;
				if (x < (w - 1)) {
					value += image[AD(x+1,y,w)+RED];
					div++;
				}
				if (x) {
					value += image[AD(x-1,y,w)+RED];
					div++;
				}
				image[AD(x,y,w)+RED] = value / div;

				div = value = 0;
				if (y < (h - 1)) {
					value += image[AD(x,y+1,w)+BLUE];
					div++;
				}
				if (y) {
					value += image[AD(x,y-1,w)+BLUE];
					div++;
				}
				image[AD(x,y,w)+BLUE] = value / div;

			} else if ( bayer == p2 ) {
				
				/* green. blue lr, red tb */
				div = value = 0;
				
				if (x < (w - 1)) {
					value += image[AD(x+1,y,w)+BLUE];
					div++;
				}
				if (x) {
					value += image[AD(x-1,y,w)+BLUE];
					div++;
				}
				image[AD(x,y,w)+BLUE] = value / div;

				div = value = 0;
				if (y < (h - 1)) {
					value += image[AD(x,y+1,w)+RED];
					div++;
				}
				if (y) {
					value += image[AD(x,y-1,w)+RED];
					div++;
				}
				image[AD(x,y,w)+RED] = value / div;

			} else {
				
				/* blue. green lrtb, red diagonals */
				div = value = 0;

				if (y) {
					value += image[AD(x,y-1,w)+GREEN];
					div++;
				}
				if (y < (h - 1)) {
					value += image[AD(x,y+1,w)+GREEN];
					div++;
				}
				if (x) {
					value += image[AD(x-1,y,w)+GREEN];
					div++;
				}
				if (x < (w - 1)) {
					value += image[AD(x+1,y,w)+GREEN];
					div++;
				}
				image[AD(x,y,w)+GREEN] = value / div;

				div = value = 0;
				if ((y < (h - 1)) && (x < (w - 1))) {
					value += image[AD(x+1,y+1,w)+RED];
					div++;
				}
				if ((y) && (x)) {
					value += image[AD(x-1,y-1,w)+RED];
					div++;
				}
				if ((y < (h - 1)) && (x)) {
					value += image[AD(x-1,y+1,w)+RED];
					div++;
				}
				if ((y) && (x < (w - 1))) {
					value += image[AD(x+1,y-1,w)+RED];
					div++;
				}
				image[AD(x,y,w)+RED] = value / div;
			}
		}

	return (0);
}

static int
gp_bayer_decode (unsigned char *input, int w, int h, unsigned char *output,
		 int tile)
{
	gp_bayer_expand (input, w, h, output, tile);
	gp_bayer_interpolate (output, w, h, tile);

	return (0);
}

static void *usbvideo_rvmalloc(unsigned long size)
{
	void *mem;
	unsigned long adr;

	size = PAGE_ALIGN(size);
	mem = vmalloc_32(size);
	if (!mem)
		return NULL;

	memset(mem, 0, size); /* Clear the ram out, no junk to the user */
	adr = (unsigned long) mem;
	while (size > 0) {
		SetPageReserved(vmalloc_to_page((void *)adr));
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}

	return mem;
}

static void usbvideo_rvfree(void *mem, unsigned long size)
{
	unsigned long adr;

	if (!mem)
		return;

	adr = (unsigned long) mem;
	while ((long) size > 0) {
		ClearPageReserved(vmalloc_to_page((void *)adr));
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}
	vfree(mem);
}

 unsigned long usbvideo_kvirt_to_pa(unsigned long adr)
{
	unsigned long kva, ret;

	kva = (unsigned long) page_address(vmalloc_to_page((void *)adr));
	kva |= adr & (PAGE_SIZE-1); /* restore the offset */
	ret = __pa(kva);
	return ret;
}



